open Num
open Nvbindlib

type term =
    FVar of term variable
  | App of term * term
  | Lambda of (term, term) binder
  | Crochet of term * stack
  | Mu of (stack, term) binder
  | Int of num
  | Extern1 of (num -> term)
  | Extern2 of (num -> num -> term)
  | Fix of (term, term) binder


and stack =
  | MuVar of stack variable
  | Push of term * stack
  | Empty
 
let rec reduce t s =
  match t,s with
  | App(u,v), s -> reduce u (Push(v,s))
  | Lambda(f), Push(v,s) -> reduce (subst f v) s
  | Crochet(u,s), _ -> reduce u s
  | Mu(f), s -> reduce (subst f s) s
  | Extern1 f, Push(n,s) ->
      (match reduce n Empty with
	Int n, Empty -> reduce (f n) s
      | _ -> t,s)
  | Extern2 f, Push(n,Push(m,s)) -> 
      (match reduce n Empty, reduce m Empty with
	(Int n, Empty),(Int m, Empty) -> reduce (f n m) s
      | _ -> t,s)
  | Fix f, _ -> reduce (subst f t) s
  | _ -> t,s

let print = function
    Int n, _ -> print_string (string_of_num n); print_newline ()
  | _ -> ()

let fVar(x) = FVar(x)

let muVar(x) = MuVar(x)

let idt = unbox (Lambda (^ bind fVar x in x ^) )

let ttrue = unbox (Lambda(^ bind fVar x in Lambda(^ bind fVar y in x ^) ^) )

let tfalse = unbox (Lambda(^ bind fVar x in Lambda(^ bind fVar y in y ^) ^) )

let zero = Int (Num.Int 0)

let one = Int (Num.Int 1)

let succ = Extern1 (fun n -> Int(n +/ (Num.Int 1)))

let pred = Extern1 (fun n -> Int(n -/ (Num.Int 1)))

let add = Extern2 (fun n m -> Int(n+/m))

let mul = Extern2 (fun n m -> Int(n*/m))

let is_null = Extern1 (fun n -> if n =/ (Num.Int 0) then ttrue else tfalse)

let fact = unbox (Fix(^ bind fVar fact in Lambda(^ bind fVar n in 
  App(^App(^App(^ (^is_null^),n^) ,(^one^) ^),App(^App(^ (^mul^) ,n^),App(^fact,App(^ (^pred^) ,n^) ^) ^) ^) ^) ^) )
 
let r = print (reduce (App(fact, Int (Num.Int 1000))) Empty)

